#ifndef _FIXMAP_H_
#define _FIXMAP_H_

/*
 * The kernel needs to be able to access some specific physical
 * pages, such as the local APIC for enabling SMP. Low physical
 * addresses can be accessed directly using the __pa macro because
 * they are mapped into the kernel's address space. Other pages
 * (like the lapic page) exist in high memory, and therefore
 * cannot be directly addressed using the default kernel page tables.
 * Instead, we create special mappings to these fixed addresses.
 */

/* adapted from Linux 2.6.11 */

#include "param.h"
#include "page.h"

enum fixed_addresses {
    __fix_no_map,
    FIX_APIC_BASE,
    FIX_IO_APIC_BASE,
    __end_of_fixed_addresses
};

/* use high virtual addresses for fixed mapped pages */
#define FIXADDR_TOP ((unsigned long)0xfffff000)
#define FIXADDR_SIZE (__end_of_fixed_addresses << PAGE_SHIFT)
#define FIXADDR_START (FIXADDR_TOP - FIXADDR_SIZE)

#define __fix_to_virt(x)    (FIXADDR_TOP - ((x) << PAGE_SHIFT))
#define __virt_to_fix(x)    ((FIXADDR_TOP -((x)&PAGE_MASK)) >> PAGE_SHIFT)

extern void __set_fixmap(enum fixed_addresses idx,
        unsigned long paddr, pgprot_t flags);

#define set_fixmap(idx, paddr) __set_fixmap(idx, paddr, PAGE_KERNEL)

extern void __this_fixmap_does_not_exist(void); /* function does not exist */

static inline unsigned long fix_to_virt(const unsigned int idx)
{
    /*
     * This is a cute static safety check borrowed from Linux. The compiler
     * will remove this conditional if idx is correct; if it is not, there is
     * no implementation of __this_fixmap..., so the linker will throw a fit.
     * One has to assume that historically in Linux people messed up indexing
     * of fixed maps all the time.
     */
    if(idx >= __end_of_fixed_addresses) 
        __this_fixmap_does_not_exist();

    return __fix_to_virt(idx);
}

static inline unsigned long virt_to_fix(const unsigned long vaddr)
{
    if(vaddr >= FIXADDR_TOP || vaddr < FIXADDR_SIZE)
        panic("bad va to virt_to_fix");
    return __virt_to_fix(vaddr);
}

#endif
