/* 
 * Physical memory allocator
 */

#include "types.h"
#include "defs.h"
#include "param.h"
#include "spinlock.h"
#include "page.h"
#include "proc.h"
#include "x86.h"

struct spinlock kalloc_lock;

/* Page descriptors */
struct page * __pages;

/* Free memory for the buddy memory allocator */
struct buddy_area mem_areas[MAX_MEM_ORDER];

/*
 * Overloads the page.plist to store the buddy area for 
 * pages allocated in blocks from kmalloc.
 */
#define SET_PAGE_AREA(pg,area) ((pg)->plist.next = (struct list_head *)(area))
#define GET_PAGE_AREA(pg)     ((struct buddy_area *)(pg)->plist.next)

void __kfree_pages(struct page* p, int order);

/*
 * Initialize free list of physical pages.
 * This code uses all of the memory from the end of
 * the kernel page table (kernel_page_table_end)
 * up to the system memory size (MEMORY_SIZE in GB from param.h)
 * (minus the pages necessary to store the `struct page' array.
 * Real systems would determine the amount of memory
 * available and use it all.
 */ 
void
init_free_lists()
{
    int order;
    for(order=0;order<MAX_MEM_ORDER;++order) {
        INIT_LIST_HEAD(&mem_areas[order].free_list);
        mem_areas[order].nr_free = 0;
    }
}

void
dump_freelists()
{
    int order;
    unsigned total_free = 0;
    for(order=0;order<MAX_MEM_ORDER;++order) {
        cprintf("mem_areas[%d].nr_free = %x\n",order,mem_areas[order].nr_free);
        total_free += mem_areas[order].nr_free*(1<<order);
    }
    cprintf("total free: %d pages\n",total_free);
}
void
kinit(void)
{
  unsigned long allocatable_mem, allocatable_pages;
  unsigned pfn;
  extern unsigned long kernel_page_table_end;
  unsigned long max_mem = MEMORY_SIZE * 1024 * 1024;
  unsigned max_pfn = max_mem >> PAGE_SHIFT;
  struct page * pg;
  int order;

  initlock(&kalloc_lock, "kalloc");

  /* total allocatable memory is everything after the kernel page
     table, excluding the accounting structure pages[] */
/*
  allocatable_mem = (unsigned)
    (max_mem - __pa(kernel_page_table_end)) *
    ((sizeof(struct page) + PAGE_SIZE) / (2.0*sizeof(struct page)+PAGE_SIZE)); */

  allocatable_mem =
    max_mem - __pa(kernel_page_table_end) -
        sizeof(struct page)*(max_mem/(double)(sizeof(struct page)+PAGE_SIZE));

  allocatable_mem = allocatable_mem & PAGE_MASK;
  allocatable_pages = allocatable_mem >> PAGE_SHIFT;

  /* init pages array and zero memory */
  __pages = (struct page *)kernel_page_table_end;

  memset(__pages,0,allocatable_mem);

  /* set up the free lists */
  init_free_lists();

  /* initialize the page descriptors */
  for( pfn = max_pfn - allocatable_pages, pg = __pages + pfn;
       pfn < max_pfn; ++pfn, ++pg)
  {
    set_page_count(pg,0);
    set_page_mapcount(pg,0);
    __kfree_pages(pg,0);
  }

  cprintf("mem = %d (%d pages)\n", allocatable_mem,allocatable_pages);
  for(order=0;order<MAX_MEM_ORDER;++order)
    cprintf("  mem_areas[%d]: %d blocks\n",order,mem_areas[order].nr_free);
}

/**** end initialization ****/


/**** Low level buddy allocator page allocation routines ****/

/*
 * Break down a large block of pages
 */
static inline void 
split_block(struct page * page, int req_order, int found_order)
{
    struct buddy_area * area;
    unsigned long size;
    int curr;

    /* Starting with the whole block that was found,
       subdivide from the upper halves until we get the block
       size requested */

    size = 1 << found_order;
    curr = found_order;
    area = &mem_areas[curr];
    while(curr > req_order) {
        area--;
        curr--;
        size >>= 1;
   
        /* grab a page directory entry in the middle of
           the block of order curr+1 */  
        list_add(&page[size].plist, &area->free_list);
        area->nr_free++;
        page[size].order = curr;
        set_bit(PG_Buddy,&page[size].flags);
    }
}

inline void __adjust_counts(struct page * pg, int cnt, int delta)
{
    int i,old;
    for(i=0;i<cnt;++i) {
        old = get_page_count(pg);
        set_page_count(pg,old + delta);
        old = get_page_mapcount(pg);
        set_page_mapcount(pg,old+delta);
        ++pg;
    }
}

/*
 * Allocate a contiguous block of 1 << order pages and return the
 * descriptor of the first page in the block. If a block of the 
 * requested order is not available, will break up an existing
 * block of higher order.
 */
struct page *
__kalloc_pages(int order)
{
    struct buddy_area * area;
    struct page * ret;
    int curr;

    acquire(&kalloc_lock);

    for(curr = order; curr < MAX_MEM_ORDER; ++curr) {
        area = &mem_areas[curr]; 
        if(list_empty(&area->free_list)) {
            /* XXX debugging sanity check */ 
            if(area->nr_free != 0) {
                cprintf("Empty free list of order %d with nr_free = %d\n",
                    curr,area->nr_free);
                panic("");;
            }
            continue;
        }

        /* Pull a block off of the free list */
        ret = list_entry(area->free_list.next, struct page, plist);
        list_del(&ret->plist);
        ret->order = 0;
        clear_bit(PG_Buddy,&ret->flags);
        area->nr_free--;

        /* If the block is larger than requested, split it and
           put the remaining pages back on the appropriate-size
           free lists */
        if(curr > order)
            split_block(ret,order,curr);

        goto exit;
    }
 bad:
    cprintf("OOM!\n");
    ret = 0;
 exit:
    release(&kalloc_lock);
    return ret;
}

/*
 * Free a contiguous block of 1 << order pages starting with
 * the decriptor p. 
 */
void
__kfree_pages(struct page* p, int order)
{
    int i,pcnt;
    unsigned long page_index;

    acquire(&kalloc_lock);

    pcnt = 1 << order;

    /* 
     * Sanity check the reference counts on all the pages
     * of this block
     */ 
    for(i=0;i<pcnt;++i) {
        struct page * check = p + i;
        if(get_page_count(check) != 0) {
            cprintf("Bad page free of frame %d, count: %d\n",
                page_to_pfn(check),get_page_count(check));
            panic("");
        }
    }    

    page_index = page_to_pfn(p);

    /*
     * Coallesce this block with its buddy, iterating
     * up through all of the sizes of blocks.
     */
    while(order < MAX_MEM_ORDER-1) {
        struct page * buddy;
        unsigned long buddy_index;

        /* Check whether the current block's buddy is free.
           The buddy is the first page of the adjacent block;
           i.e., the page found by flipping the order'th bit of
           the page frame number */

        buddy_index = page_index ^ (1<<order);
        buddy = pfn_to_page(buddy_index);
       
        /* conditions: free and has correct order; if not, 
           we give up coallescing */
        if(get_page_count(buddy) != 0 || 
           !test_bit(PG_Buddy,&buddy->flags) ||
           buddy->order != order)
            break;

        /* pull buddy off of current order's free list */
        list_del(&buddy->plist);
        mem_areas[order].nr_free--;
       
        /* coallesce the page with its buddy and bump up to the
           next level of block size the start of the new coallesced block is
           the lower page of the two original blocks */
        buddy->order = 0;
        clear_bit(PG_Buddy,&buddy->flags);
        page_index &= buddy_index;
        ++order; 
    }

    /* Set up the final coallesced block on the appropriate free list */
    p = pfn_to_page(page_index);
    p->order = order;
    set_bit(PG_Buddy,&p->flags);
    list_add(&p->plist,&mem_areas[order].free_list);
    mem_areas[order].nr_free++;

 exit:
    release(&kalloc_lock);    
}


/**** end buddy allocator ****/

/*
 * Allocate anonymous memory for the kernel. Returns a kernel
 * VA for the allocated memory.
 *
 * NB Currently using a page-granularity buddy system allocator;
 *    minimum allocation is thus 4096 bytes, so massive internal
 *    fragmentation is the rule.
 *
 *    The maximum request size under this current allocator is
 *    thus 4096 * 2 << 10 == 4MB
 */
void *
kmalloc(uint size)
{
    struct page * pg;
    void * vaddr;
    int order;

    for(order=0;order<MAX_MEM_ORDER;++order) {
        if(size <= (PAGE_SIZE * (1UL<<order))) {
            pg = __kalloc_pages(order);
            goto allocated;
        }
    }
    cprintf("kmalloc(%ld) exceeds maximum request size of 4MB\n",size);
    panic("");

 allocated:
    /* don't panic */
    /* if(!pg)
        panic("Failed to allocate kernel anonymous page\n"); */
    if(!pg)
        return 0;

    /* expensive zeroing */
    vaddr = page_to_virt(pg);
    memset(vaddr,0,(1<<order));

    /* accounting for kfree */
    SET_PAGE_AREA(pg,&mem_areas[order]);
    __adjust_counts(pg,1<<order,1);

    return __va(page_to_pfn(pg) << PAGE_SHIFT);
}

/*
 * For virtual addresses allocated with kmalloc().
 *
 * Returns the number of pages freed.
 */
int
kfree(void * addr)
{
    struct page * pg;
    struct buddy_area * area;
    int order;

    pg = pfn_to_page(__pa(addr) >> PAGE_SHIFT);
    area = GET_PAGE_AREA(pg);

    order = (int)(area - mem_areas);

    /* sanity */
    if(order < 0 || order >= MAX_MEM_ORDER) {
        cprintf("kfree(%x) frees block not allocated by kmalloc\n",addr);
        panic("");
    }

    __adjust_counts(pg,1<<order,-1);
    __kfree_pages(pg,order);
    
    return 1<<order;
}

static void
map_user_page(struct proc *p, struct page * pg, unsigned long vaddr)
{
    pte_t * pte;
    pgd_t * pgd;
    
    pgd = &p->mm.pgd[pgd_index(vaddr)];
    if(pgd_none(*pgd)) {
        pte = (pte_t*)ualloc_page_table(p);
        set_pgd(pgd, __pgd(__pa(pte) | _PAGE_TABLE));
    } else {
        pte = (pte_t*) __va((pgd_pfn(*pgd) << PAGE_SHIFT));
    }

    set_pte(&pte[pte_index(vaddr)], pfn_pte(page_to_pfn(pg),PAGE_USERPROC));
}

/*
 * Get a page for a page table for a process
 */
void *  
ualloc_page_table(struct proc*p)
{
    pte_t * pte;

    pte = (pte_t*)kmalloc(PAGE_SIZE);
    if(0 != pte) {
        memset(pte,0,PAGE_SIZE);
        p->mm.mm_hist.pt_alloc_cnt += 1;
    }
    return pte;
}

void
ufree_page_table(struct proc*p, uint pfn)
{
    kfree(__va(pfn << PAGE_SHIFT));
    p->mm.mm_hist.pt_free_cnt += 1;
}


/*
 * Allocate user memory at a particular address (the amount
 * allocated is rouned up to page boundary)
 *
 * Question: As a consequence of the no-management-structure
 * implementation, each page is allocated individually. A more
 * reasonable approach would be to use kalloc() to allocate
 * larger blocks of memory. What additional data structures
 * are necessary to support management (tracking and freeing)
 * of memory allocated in this way?
 *
 */
int
ualloc(struct proc *p, void * addr, int sz)
{
    int np;
    unsigned long vaddr = (unsigned long)addr;
    struct page * pg;

    void * newalloc;

    if(sz % PAGE_SIZE)
        sz = (sz + PAGE_SIZE) & PAGE_MASK;
    np = sz >> PAGE_SHIFT;

    while(np>0) {
        newalloc = kmalloc(PAGE_SIZE);
        if(!newalloc) {
            cprintf("OOM allocating %d bytes at %p in proc %d\n",
                    sz,addr,p->pid);
                    return -1;
            return -1;
        }
        pg = virt_to_page(newalloc);
        map_user_page(p,pg,vaddr);
        vaddr += PAGE_SIZE;

        p->mm.mm_hist.alloc_cnt += 1;
        np -= 1;
    }

    // kick tlb
    if(p == cp)
        flush_tlb();

    return 0;
}

/* 
 * Free the range of memory [addr,addr+sz) allocated
 * by ualloc(). 
 */
void
ufree(struct proc *p, void * addr, int sz)
{
    struct page * pg;
    pte_t * pte;
    int i,j,np,freed;
    if(sz % PAGE_SIZE)
        sz = (sz + PAGE_SIZE) & PAGE_MASK;
    np = sz >> PAGE_SHIFT;

    for(j=0;j<np;) {
        // free memory
        pte = walk_to_pte(p->mm.pgd,addr);
        if(!pte || pte_none(*pte)) {
            cprintf("Attempt to free unmapped process memory at %x\n",addr);
            return;
        }
        pg = pfn_to_page(pte_pfn(*pte));

        freed = kfree(__va(pte_pfn(*pte) << PAGE_SHIFT));

        p->mm.mm_hist.free_cnt += freed;

        // clear page table entries
        for(i=0;i<freed;++i) {
            pte = walk_to_pte(p->mm.pgd,addr);
            if(pte && !pte_none(*pte)) {
                set_pte(pte,__pte(0)); 
            }
            addr += PAGE_SIZE; 
        }
        j += freed;
    }

    // whack tlb
    if(p == cp)
        flush_tlb();
}

/*
 * Remap a page range for a user process
 */
int
uremap(struct proc * p, void * orig, void * new, int sz)
{
    int pfns, i;
    unsigned long oaddr = (unsigned long)orig;
    unsigned long naddr = (unsigned long)new;
    pte_t *opte, *npte;

    if(sz % PAGE_SIZE)
        sz = (sz + PAGE_SIZE) & PAGE_MASK;

    pfns = sz >> PAGE_SHIFT;
    for(i=0;i<pfns;++i) {
        opte = walk_to_pte(p->mm.pgd,(void*)oaddr);
        npte = walk_to_pte(p->mm.pgd,(void*)naddr);

        if(!opte || !npte) {
            cprintf("Error, remapping oaddr %x (%p) naddr %x (%p)\n",
                oaddr,opte,naddr,npte);
            return -1;
        }

        set_pte(npte,*opte); 
        set_pte(opte,__pte(0));

        oaddr += PAGE_SIZE;
        naddr += PAGE_SIZE;            
    }

    // kick tlb
    flush_tlb();

    return 0;
}
