#include "types.h"
#include "defs.h"
#include "param.h"
#include "x86.h"
#include "mmu.h"
#include "fixmap.h"
#include "page.h"

/* kernel page directory */
extern pgd_t kernel_pgd[];

unsigned int kernel_page_table_end; /* will hold start of "free" memory */

/* post-paging GDT */
struct segdesc ppgdt[] = 
{
    SEG_NULL,
    
    // kernel code segment
    [GD_KT >> 3] = SEG(STA_X|STA_R, 0x0, 0xffffffff, 0),

    // kernel data segment
    [GD_KD >> 3] = SEG(STA_W, 0x0, 0xffffffff, 0),

    // user code
    [GD_UT >> 3] = SEG(STA_X | STA_R, 0x0, 0xffffffff, 3),
    
    // user data
    [GD_UD >> 3] = SEG(STA_W, 0x0, 0xffffffff, 3),

    // tss?
    [GD_TSS >> 3] = SEG_NULL
};

static unsigned int emulate_va2pa(pgd_t * pgdir, void * va);

static void reserve_fixmap_mem(pgd_t * pgdir, void ** pg0);

/*
 * Initialize the kernel page tables.
 * 
 * Claims the up to the initial 1MB of pages immediately following _end
 * for page tables. Will fill in kernel_page_table_end with the
 * next free location.
 *
 * NB we know in advance how many pages this will actually require based
 *    on the MEMORY_SIZE parameter, but we'll pretend we don't as it's
 *    as simple.
 */

void paging_init(void)
{
    extern int end;
    unsigned int pgd_idx, pte_offset;
    unsigned int pfn;
    unsigned long cr0;
    void * pg0;
    pte_t * pte;
    pgd_t * pgd;

    unsigned int max_pfn = (MEMORY_SIZE * 1024 * 1024) >> PAGE_SHIFT;

    /* avoid nasty surprises */
    memset(kernel_pgd,0,PGD_PTRS*sizeof(pgd_t));

    /* pg0 holds the next available page of our scratch space */
    pg0 = &end;
    pg0 = (void*)(((uint)pg0 + PAGE_SIZE) & PAGE_MASK);

    /* populating the page directory for [KERNEL_OFFSET, ...) */
    pgd = &kernel_pgd[pgd_index(KERNEL_OFFSET)];

    cprintf("paging_init:\n"
            "\tkernel pgd: %p\n"
            "\tend: %p\n"
            "\tpg0: %p\n" 
            "\tpgd index: %d\n"
            "\tmax_pfn: %d\n",
        kernel_pgd,&end,pg0,pgd_index(KERNEL_OFFSET),max_pfn);

    pgd_idx = pgd_index(KERNEL_OFFSET);
    pfn = 0;
    for(; pgd_idx < PGD_PTRS && pfn < max_pfn; ++pgd_idx, ++pgd) {
        /* wipe page table */
        memset(pg0,0,PGT_PTRS*sizeof(pte_t));

        /* set up page directory entry */
        set_pgd(pgd,__pgd(__pa(pg0) | _PAGE_TABLE));

        pte = (pte_t*)pg0;
        for(pte_offset=0;pte_offset < PGT_PTRS && pfn < max_pfn; 
            ++pte_offset,++pte,++pfn)
        {
            set_pte(pte,pfn_pte(pfn,PAGE_KERNEL));
        }

        pg0 =(void*)(((uint)pg0 + PAGE_SIZE) & PAGE_MASK);
    }

    /* Also set up fixed mappings */
    reserve_fixmap_mem(kernel_pgd,&pg0);

    /* record the next free physical page */
    kernel_page_table_end = (unsigned long)pg0;

    /* create identity map for the first 4MB */
    kernel_pgd[0] = kernel_pgd[pgd_index(KERNEL_OFFSET)];

    /* turn on paging */
    write_cr3(__pa(kernel_pgd));

    cr0 = read_cr0();
    cr0 |= CR0_PE|CR0_PG|CR0_AM|CR0_WP|CR0_NE|CR0_TS|CR0_EM|CR0_MP; 
    cr0 &= ~(CR0_TS|CR0_EM);
    write_cr0(cr0);

    /* now fix up the segment registers to rationalize LA and VA mapping */
    lgdt(ppgdt,sizeof(ppgdt));
    __asm__ __volatile__("movw %%ax,%%gs" :: "a" (GD_UD|3));
    __asm__ __volatile__("movw %%ax,%%fs" :: "a" (GD_UD|3));
    __asm__ __volatile__("movw %%ax,%%es" :: "a" (GD_KD));
    __asm__ __volatile__("movw %%ax,%%ds" :: "a" (GD_KD));
    __asm__ __volatile__("movw %%ax,%%ss" :: "a" (GD_KD));
    __asm__ __volatile__("ljmp %0,$1f\n 1:\n" :: "i" (GD_KT));
    __asm__ __volatile__("lldt %%ax" :: "a" (0));

    /* whack the identity mappint */
    kernel_pgd[0] = __pgd(0);

    /* todo: set up page tables for fixed mapped pages */

    flush_tlb(); 
}

/* XXX assumes less than 4MB of fixed mapped pages */
static void
reserve_fixmap_mem(pgd_t * pgdir, void ** pg0)
{
    pgd_t * pgd;
    cprintf("Initializting fixmaps\n");

    pgd = &pgdir[pgd_index(FIXADDR_START)];

    /* set up page directory entry */
    set_pgd(pgd,__pgd(__pa(*pg0) | _PAGE_TABLE));

    /* wipe page table */
    memset(*pg0,0,PGT_PTRS*sizeof(pte_t));
    
    *pg0 =(void*)(((uint)*pg0 + PAGE_SIZE) & PAGE_MASK);
}

/* 
 * Emulate the hardware translation from virtual to physical
 * addressing for testing.
 */
static unsigned int
emulate_va2pa(pgd_t * pgdir, void * va)
{
    pte_t * pte;
    
    pgdir = &pgdir[pgd_index((unsigned int)va)];
    if(!(pgd_val(*pgdir) & _PAGE_PRESENT))
        return ~0; 
    pte = (pte_t*) __va((pgd_pfn(*pgdir) << PAGE_SHIFT));
    if(!(pte_val(pte[pte_index((unsigned int)va)]) & _PAGE_PRESENT))
        return ~0;
    return pte_pfn(pte[pte_index((unsigned int)va)]) << PAGE_SHIFT;
}

void __set_fixmap(enum fixed_addresses idx,
    unsigned long paddr, pgprot_t flags)
{
    pgd_t * pgd;
    pte_t * pte;
    unsigned long pfn = paddr >> PAGE_SHIFT;
    unsigned long addr = __fix_to_virt(idx);

    if(idx >= __end_of_fixed_addresses)
        panic("bad index to __set_fixmap");

    pgd = &kernel_pgd[pgd_index(addr)];
    if(pgd_none(*pgd))
        panic("invalid page directory entry mapping fixed address");
    
    pte = (pte_t*) __va((pgd_pfn(*pgd) << PAGE_SHIFT));
    if(pgd_none(*pgd))
        panic("invalid page table entry mapping fixed address");

    set_pte(&pte[pte_index(addr)], pfn_pte(pfn,flags));

    flush_tlb_one(addr);
}

pte_t *
walk_to_pte(pgd_t * pgdir, void * va)
{
    pte_t * pte;
    
    pgdir = &pgdir[pgd_index((unsigned int)va)];
    if(pgd_none(*pgdir)) {
        pte = (pte_t*)kmalloc(PAGE_SIZE);
        memset(pte,0,PAGE_SIZE);
        set_pgd(pgdir, __pgd(__pa(pte) | _PAGE_TABLE));
    } else {
        pte = (pte_t*) __va(pgd_pfn(*pgdir) << PAGE_SHIFT);
    }
    pte = &pte[pte_index((unsigned int)va)];

    return pte;
}

/*
 * Prints the non-zero entries in the page tables
 */
void
dump_page_table(pgd_t * pgdir)
{
    unsigned long addr;
    int pgd_idx, pt_idx;
    pte_t * pte;

    for(pgd_idx=0;pgd_idx < PGD_PTRS;++pgd_idx,++pgdir) {
        if(pgd_none(*pgdir))
            continue;

        // skip kernel maps
        if(pgd_idx >= pgd_index(KERNEL_OFFSET))
            continue;

        cprintf("[%d]\n",pgd_idx);
        pte = (pte_t*) __va(pgd_pfn(*pgdir) << PAGE_SHIFT);
        for(pt_idx=0;pt_idx<PGT_PTRS;++pt_idx,++pte) {
            if(pte_none(*pte))
                continue;

            addr = pgd_idx << PGD_SHIFT;
            addr |= pt_idx << PAGE_SHIFT;

            cprintf("%x -> %d\n",addr,pte_pfn(*pte));
        }
    }
}

void
copy_kernel_maps(pgd_t *dst)
{
    pgd_t * kpgd = kernel_pgd;
    pgd_t * pgd = dst;
    int pgdir_index=0;
    for( ; pgdir_index < PGD_PTRS; ++pgdir_index,++kpgd,++pgd) {
        set_pgd(pgd,*kpgd); 
    }
}
