#ifndef _PAGE_H_
#define _PAGE_H_

#include "param.h"
#include "list.h"

/* 2-level page table (4KB pages), 32 bit addressing */

#define PAGE_SHIFT      12
#define PAGE_SIZE       (1 << PAGE_SHIFT)
#define PAGE_MASK       (~(PAGE_SIZE-1))

#define PGT_PTRS        1024  /* PTEs per page table */

#define PGD_SHIFT       22
#define PGD_PTRS        1024

/* Flags for PTEs */
#define _PAGE_PRESENT   0x001
#define _PAGE_RW    0x002
#define _PAGE_USER  0x004
#define _PAGE_PWT   0x008
#define _PAGE_PCD   0x010
#define _PAGE_ACCESSED  0x020
#define _PAGE_DIRTY 0x040
#define _PAGE_PSE   0x080   
#define _PAGE_GLOBAL    0x100  
#define _PAGE_UNUSED1   0x200  
#define _PAGE_UNUSED2   0x400
#define _PAGE_UNUSED3   0x800
#define _PAGE_NX    0 /* no PAE mode */

/* default flags for page directory entries that point to page tables */
//#define _PAGE_TABLE (_PAGE_PRESENT | _PAGE_RW | _PAGE_USER | _PAGE_ACCESSED | _PAGE_DIRTY)
#define _PAGE_TABLE (_PAGE_PRESENT | _PAGE_RW | _PAGE_USER )
#define _PAGE_KERNEL \
    (_PAGE_PRESENT | _PAGE_RW | _PAGE_DIRTY | _PAGE_ACCESSED | _PAGE_NX)
#define _PAGE_KERNEL_EXEC \
    (_PAGE_PRESENT | _PAGE_RW | _PAGE_DIRTY | _PAGE_ACCESSED)

#define _PAGE_USERPROC \
    (_PAGE_PRESENT | _PAGE_RW | _PAGE_DIRTY | _PAGE_ACCESSED | _PAGE_USER)

/* move between physical and virtual in kernel mode */
#define __pa(x) ((unsigned long)(x)-KERNEL_OFFSET)
#define __va(x) ((void*)((unsigned long)(x)+KERNEL_OFFSET))

/* 
 * Using these accessor and set methods assigns types to
 * what would otherwise be a lot of untyped operation and
 * (hopefully) helps us stay sane.
 */
#ifndef __ASSEMBLER__
typedef struct { unsigned long pte; } pte_t;
typedef struct { unsigned long pgd; } pgd_t;
typedef struct { unsigned long pgprot; } pgprot_t;
#endif

#define PAGE_KERNEL     __pgprot(_PAGE_KERNEL)
#define PAGE_KERNEL_EXEC    __pgprot(_PAGE_KERNEL_EXEC)
#define PAGE_USERPROC   __pgprot(_PAGE_USERPROC)

#define __pte(x) ((pte_t) { (x) } )
#define __pgd(x) ((pgd_t) { (x) } )
#define __pgprot(x) ((pgprot_t) { (x) } )
#define pgd_val(x) ((x).pgd)
#define pte_val(x) ((x).pte)
#define pgprot_val(x)   ((x).pgprot)

#define set_pgd(pgdptr, pgdval) (*(pgdptr) = pgdval)
#define set_pte(pteptr, pteval) (*(pteptr) = pteval)
#define pfn_pte(pfn, prot)  __pte(((pfn) << PAGE_SHIFT) | pgprot_val(prot))

#define pgd_pfn(x) ((unsigned long)((x).pgd >> PAGE_SHIFT))
#define pte_pfn(x) ((unsigned long)((x).pte >> PAGE_SHIFT))

#define pte_none(x)     (!(x).pte)
#define pgd_none(x)     (!(x).pgd)

/* accessing frame number or flags from a pde or pte */
#define PTE_PFN_MASK        ((pte_t)PAGE_MASK)
#define PTE_FLAGS_MASK      (~PTE_PFN_MASK)

#ifndef __ASSEMBLER__
/* Find the page table entry for a virtual address */
pte_t * walk_to_pte(pgd_t * pgdir, void * addr);
/* debugging */
void dump_page_table(pgd_t * pgdir);
#endif

/* index into page directory */
#define pgd_index(addr) (((addr) >> PGD_SHIFT) & (PGD_PTRS-1))
#define pte_index(addr) (((addr) >> PAGE_SHIFT) & (PGT_PTRS-1))

#ifndef __ASSEMBLER__
extern struct page * __pages;
#endif

#define pfn_to_page(pfn) (__pages + (pfn))
#define page_to_pfn(page) ((unsigned long)((page) - __pages))
/* only works on kernel virtual addresses */
#define virt_to_page(kaddr) pfn_to_page(__pa(kaddr) >> PAGE_SHIFT)
#define page_to_virt(page)  (__va(page_to_pfn(page) << PAGE_SHIFT))

#define PG_Buddy    1

#ifndef __ASSEMBLER__
struct page {
    unsigned long flags;        /* various flags */
    int count;                  /* process reference count-1: -1 means free */
    int mapcount;               /* page mapping count-1; always >= count */

    unsigned char order;        /* allocation order for page allocator */
                                /* only valid if the PG_Buddy flag is set */

    struct list_head plist;     /* handle for placing on free list, etc. */
};

struct buddy_area {
    int nr_free;
    struct list_head free_list;     /* free memory */
};

static inline void set_page_count(struct page * pg, int cnt)
{
    pg->count = cnt-1;
}
static inline int get_page_count(struct page * pg)
{
    return pg->count+1;
}
static inline void set_page_mapcount(struct page * pg, int cnt)
{
    pg->mapcount = cnt-1;
}
static inline int get_page_mapcount(struct page * pg)
{
    return pg->mapcount+1;
}

/** misc. utilities **/

#define flush_tlb_one(addr) \
    __asm__ __volatile__("invlpg %0": :"m" (*(char*)(addr)))
static inline void write_cr3(unsigned long cr3)
{
    __asm__ __volatile__ ("mov %0, %%cr3" : : "r" (cr3) : "memory");
}

static inline unsigned long read_cr3(void)
{
    unsigned long cr3;
    __asm__ __volatile__ ( "mov %%cr3, %0" : "=r" (cr3) : );
    return cr3;
}

static inline void flush_tlb(void)
{
    write_cr3(read_cr3());
}

static inline unsigned long read_cr0(void)
{
    unsigned long cr0;
    __asm__ __volatile__ ( "mov %%cr0,%0" : "=r" (cr0));
    return cr0;
}

static inline void write_cr0(unsigned long cr0)
{
    __asm__ __volatile__ ( "mov %0,%%cr0" : : "r" (cr0));
}

#endif

#endif
